
{$G+}
unit playzsm6; { Mit FX } { generelle Umstellung auf 4 Bit Delta Samples }

interface      { mit Message-Block }


function load_zsm(name: string): byte; (* Laedt ein ZSTM *)
procedure clear_zsm; (* Gibt den Speicher wieder frei und setzt Werte auf 0 *)
procedure goto_sequ(num: word);
procedure read_row;
procedure init_zsm_soundbuffer(size: word);
procedure fill_soundbuffer;
procedure calc_pattern_sizes;
procedure inc_sequ;
procedure dec_sequ;
procedure zsm_playfx(sample: byte; freq: word; vol: byte);
procedure zsm_loadfx(sample: byte; n: string);
procedure adapt_zsm_pitchtables;
procedure get_playing_time;
procedure get_maxtracks;


function patternpos: word;

type zsm_header = record
  kennung: longint; (* $4D54535A "ZSTM" *)
  version: byte;
  titel: array[0..29] of char;
  composer: array[0..19] of char;
  date_year: word;
  date_month: byte;
  date_day: byte;
  sample_entries: byte;
  sequencer_entries: word;
  sequencer_loopstart: word;
  sequencer_loopend: word;
  patterndata_size: word;
  cmsg_size: word;
end;

type smp_info = record
  length: word;
  loopstart: word;
  loopst_phase: byte;
  marker: byte;
end;

type seq_time_pos_struc = array[1..1024] of longint;



var
  qfactor: byte; { 1 = 11025 Hz,  2 = 22050 Hz,  4 = 44100 Hz }

  playing_time: longint;

  loopcount: word;
  smp_playing: array[1..63] of byte;
  smp_trig: array[1..63] of byte;
  smp_freq: array[1..63] of byte;
  smp_freq_fine: array[1..63] of longint;
  smp_pos_show: array[1..63] of word;
  smp_instances: array[1..63] of byte;
  channel_smp: array[0..15] of byte;
  channel_vol: array[0..15] of byte;
  traffic: longint;  { alles  nur fr Player }


  cmsg_pointer: pointer;

  pattheaderandtables: word;
  patternrowstotal: byte;
  patternrow_act: byte;
  chan_event: array[0..15] of bytebool;
  chan_trig: array[0..15] of byte;
  ptablesize: word;
  smpblocksize: longint;

  zsm_realsamples: byte;

  actual_sequ: word;
  zsmheader: zsm_header;

  offset_in_patterndata: word;
  smallest_pattern: word;
  biggest_pattern: word;
  number_of_patterns: word;

  act_pattern_number: word;
  act_pattern_size: word;

  actual_spl_per_row: word;

  mixvol: byte;

  vumeter: word;

  trk_active: array[0..15] of bytebool;

  track_select: array[0..15] of bytebool;

  zsmfilesize: longint;

  play_sample: array[0..15] of bytebool;

  emptyflagbits: word;
  purebits: word;
  {
  pattdisp: array[0..23, 0..15] of longint;
  npattline: array[0..15] of longint;
  npattline_smp: array[0..15] of byte;
  }
  maxtracks: byte;
  scope_ptr: pointer;
  scopemode: bytebool;
  sequ_time_pointer: pointer;

  sequ_time_pos: ^seq_time_pos_struc absolute sequ_time_pointer;


  sync: array[0..255] of byte;
  smpinfo: array[0..63] of smp_info;


{procedure new_pattline;}

implementation

uses sblaster, grpfile;

const deltatab: array[0..255] of word =
($0000, $0001, $0002, $0004, $0008, $0010, $0020, $0040,
$0080, $00C0, $00E0, $00F0, $00F8, $00FC, $00FE, $00FF,
$0100, $0101, $0102, $0104, $0108, $0110, $0120, $0140,
$0180, $01C0, $01E0, $01F0, $01F8, $01FC, $01FE, $01FF,
$0200, $0201, $0202, $0204, $0208, $0210, $0220, $0240,
$0280, $02C0, $02E0, $02F0, $02F8, $02FC, $02FE, $02FF,
$0400, $0401, $0402, $0404, $0408, $0410, $0420, $0440,
$0480, $04C0, $04E0, $04F0, $04F8, $04FC, $04FE, $04FF,
$0800, $0801, $0802, $0804, $0808, $0810, $0820, $0840,
$0880, $08C0, $08E0, $08F0, $08F8, $08FC, $08FE, $08FF,
$1000, $1001, $1002, $1004, $1008, $1010, $1020, $1040,
$1080, $10C0, $10E0, $10F0, $10F8, $10FC, $10FE, $10FF,
$2000, $2001, $2002, $2004, $2008, $2010, $2020, $2040,
$2080, $20C0, $20E0, $20F0, $20F8, $20FC, $20FE, $20FF,
$4000, $4001, $4002, $4004, $4008, $4010, $4020, $4040,
$4080, $40C0, $40E0, $40F0, $40F8, $40FC, $40FE, $40FF,
$8000, $8001, $8002, $8004, $8008, $8010, $8020, $8040,
$8080, $80C0, $80E0, $80F0, $80F8, $80FC, $80FE, $80FF,
$C000, $C001, $C002, $C004, $C008, $C010, $C020, $C040,
$C080, $C0C0, $C0E0, $C0F0, $C0F8, $C0FC, $C0FE, $C0FF,
$E000, $E001, $E002, $E004, $E008, $E010, $E020, $E040,
$E080, $E0C0, $E0E0, $E0F0, $E0F8, $E0FC, $E0FE, $E0FF,
$F000, $F001, $F002, $F004, $F008, $F010, $F020, $F040,
$F080, $F0C0, $F0E0, $F0F0, $F0F8, $F0FC, $F0FE, $F0FF,
$F800, $F801, $F802, $F804, $F808, $F810, $F820, $F840,
$F880, $F8C0, $F8E0, $F8F0, $F8F8, $F8FC, $F8FE, $F8FF,
$FC00, $FC01, $FC02, $FC04, $FC08, $FC10, $FC20, $FC40,
$FC80, $FCC0, $FCE0, $FCF0, $FCF8, $FCFC, $FCFE, $FCFF,
$FE00, $FE01, $FE02, $FE04, $FE08, $FE10, $FE20, $FE40,
$FE80, $FEC0, $FEE0, $FEF0, $FEF8, $FEFC, $FEFE, $FEFF,
$FF00, $FF01, $FF02, $FF04, $FF08, $FF10, $FF20, $FF40,
$FF80, $FFC0, $FFE0, $FFF0, $FFF8, $FFFC, $FFFE, $FFFF);

type fixkomma = record
  vor: byte;
  nach: word;
end;

type fixkomma_32bit = record
  vor: word;
  nach: word;
end;


type pitch_table = array[0..32 * 63 - 1] of fixkomma;

type seq_struc = array[1..1024] of word;

type byte_struc = array[0..65527] of byte;

type shortint_struc = array[0..65527] of shortint;

type integer_struc = array[0..32766] of word;

type pattern_header = record
  rows: byte;
  spl_per_row: word;
  tracks_active: word;
end;

type pattern_event = record
  smp: byte;
  fixvor: byte;
  fixnach: word;
  vol: byte;
  event: bytebool;
end;

type repeat_data = record
  smp: byte;
  fixvor: byte;
  fixnach: word;
end;


var
  bp_back: word;
  smp_pos: word;
  last_bp: word;
  di_: word;
  dh_: byte;
  si_: word;
  bx_: word;
  ax_: word;
  si_start: word;
  ebp_: longint;
  last_ah: byte;
  loopstphase: byte;
  trig: byte;
  channel_trig: array[0..15] of byte;
  channel_last_bp: array[0..15] of byte;
  channel_dh_: array[0..15] of byte;
  channel_last_ah: array[0..15] of byte;
  channel_si_: array[0..15] of word;
  channel_bx_: array[0..15] of word;
  channel_ax_: array[0..15] of word;
  channel_ebp: array[0..15] of longint;
  phase: byte;
  pitch_table_len: array [1..63] of byte;
  pitch_table_offset: array [1..63] of word;
  pitch_tables_pointer: pointer;
  pitch_tables: ^pitch_table absolute pitch_tables_pointer;
  pitch_tables_bytesize: word;
  sequencer_pointer: pointer;
  sequencer: ^seq_struc absolute sequencer_pointer;
  patterndata_pointer: pointer;
  patterndata: ^byte_struc absolute patterndata_pointer;
  sample_pointer: array[1..63] of pointer;
  smp_pointer: pointer;
  onesample: ^shortint_struc absolute smp_pointer;
  bitread_pos: longint;
  patternheader: pattern_header;
  {trk_active: array[0..7] of bytebool;}
  onlyvol_exist: array[0..15] of bytebool;
  actual_row: byte;
{  actual_sequ: word;}
{  actual_spl_per_row: word;}

  sampnr_table_len: array[0..15] of byte;
  sampnr_table: array[0..15, 0..15] of byte;
  volume_table_len: array[0..15] of byte;
  volume_table: array[0..15, 0..15] of byte;
  do_repeat: array[0..15] of bytebool;
  use_emptyrow_flags: bytebool;
  row: byte;
  channel_phase: array[0..15] of byte;
  pattern_row: array[0..15] of pattern_event;
  last_pattern_row: array[0..15] of repeat_data;
  s_pos: fixkomma_32bit;
  s_speed: fixkomma_32bit;
  smp_position: array[0..15] of fixkomma_32bit;
  smp_pos_before: array[0..15] of fixkomma_32bit;
  s_pos_before: fixkomma_32bit;
  actual_smplen: word;
  play_sample_actual: bytebool;
  actual_chan_vol: byte;
  actual_sample_loopstart: word;
  fill_length: word;

  play_sample_looped: array[0..15] of bytebool;

  spl_pos_in_row: word;

  row_countdown: word;
  mix_position: word;

  soundbuffer16bit_pointer: pointer;
  soundbuffer16bit: ^integer_struc absolute soundbuffer16bit_pointer;
  soundbuffer_size: word;
  soundbuffer8bit_pointer: pointer;


  pattern_size_table_pointer: pointer;
  pattern_size_table: ^seq_struc absolute pattern_size_table_pointer;


  fx_position: array[1..32] of fixkomma_32bit;
  fx_speed: array[1..32] of fixkomma_32bit;
  fx_length: array[1..32] of word;
  fx_vol: array[1..32] of byte;
  fx_pointer: array[1..32] of pointer;
  play_fx: array[1..32] of bytebool;

  const
    bitsneeded: array[0..256] of byte =
      (0,0,1,2,2,3,3,3,3,4,4,4,4,4,4,4,4,5,5,5,5,
      5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,
      6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
      6,6,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
      8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
      8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
      8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
      8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
      8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
      8,8,8,8,8);


procedure zsm_playfx(sample: byte; freq: word; vol: byte);
  var
    fullstep: real;
begin
  fullstep := freq / 22050;
  fx_speed[sample].vor := trunc(fullstep);
  fx_speed[sample].nach := trunc((fullstep - fx_speed[sample].vor) * 65536);
  fillchar(fx_position[sample], 4, 0); {triggern}
  play_fx[sample] := true;
  fx_vol[sample] := vol;
end;

procedure zsm_loadfx(sample: byte; n: string);
  var
    f: file;
begin
  {
  assign(f, n);
  reset(f, 1);
  fx_length[sample] := filesize(f);
  getmem(fx_pointer[sample], fx_length[sample]);
  blockread(f, fx_pointer[sample]^, fx_length[sample]);
  close(f);
  }

  open_file_grp(n);
  fx_length[sample] := filesize_grp;
  getmem(fx_pointer[sample], fx_length[sample]);
  readfile_grp(fx_pointer[sample], fx_length[sample]);
  close_grpfile;

end;



{
procedure new_pattline;
  var
    i: byte;
begin
  for i := 23 downto 1 do
    move(pattdisp[i-1, 0], pattdisp[i, 0], 64);
  move(npattline, pattdisp, 64);

end;
}


function load_zsm(name: string): byte;
var
  f: file;
  i: byte;
  {s: string;}  (* nur zum Test *)
  counter: word;
label skip;
begin

  fillchar(fx_length, 64, 0);

  load_zsm := 0;
  assign(f, name);
  {$I-}
  reset(f, 1);
  {$I+}
  if ioresult <> 0 then
  begin
    load_zsm := 1;
    goto skip;
  end;

{  maxtracks := 0;}


  zsmfilesize := 0;

  (* Header *)
  blockread(f, zsmheader, sizeof(zsmheader));
  if zsmheader.kennung <> $4D54535A then
  begin
    close(f);
    load_zsm := 2;
    goto skip;
  end;
  if zsmheader.version <> 4 then
  begin
    close(f);
    load_zsm := 3;
    goto skip;
  end;
  inc(zsmfilesize, sizeof(zsmheader));


  (* Message Block *)
  getmem(cmsg_pointer, zsmheader.cmsg_size);
  blockread(f, cmsg_pointer^, zsmheader.cmsg_size);
  inc(zsmfilesize, zsmheader.cmsg_size);


  (* Sample Laengen, Schleifen, Marker *)
  fillchar(smpinfo[0], 6, 0);
  blockread(f, smpinfo[1], zsmheader.sample_entries * 6);
  inc(zsmfilesize, zsmheader.sample_entries * 6);


  zsm_realsamples := 0;
  smpblocksize := 0;
  for i := 1 to zsmheader.sample_entries do
  begin
    inc(smpblocksize, smpinfo[i].length shr 1 + smpinfo[i].length and 1);
    if smpinfo[i].length > 0 then inc(zsm_realsamples);
  end;
  inc(zsmfilesize, smpblocksize);

  (* Pitch Tabellen *)
  blockread(f, pitch_table_len, zsmheader.sample_entries);
  inc(zsmfilesize, zsmheader.sample_entries);
  counter := 0;
  for i := 1 to zsmheader.sample_entries do
  begin
    pitch_table_offset[i] := counter div 3;
    inc(counter, pitch_table_len[i] * 3);
  end;
  pitch_tables_bytesize := counter;
  ptablesize := counter;
  getmem(pitch_tables_pointer, counter);
  blockread(f, pitch_tables^, counter);
  inc(zsmfilesize, counter);

  (* Sequencer *)
  getmem(sequencer_pointer, zsmheader.sequencer_entries * 2);

  getmem(sequ_time_pointer, zsmheader.sequencer_entries * 4); { nur Player }

  blockread(f, sequencer^, zsmheader.sequencer_entries * 2);
  inc(zsmfilesize, zsmheader.sequencer_entries * 2);

  if zsmheader.sequencer_loopstart < 1 then zsmheader.sequencer_loopstart
    := 1;
  if zsmheader.sequencer_loopstart > zsmheader.sequencer_entries then
    zsmheader.sequencer_loopstart := zsmheader.sequencer_entries;

  if zsmheader.sequencer_loopend < zsmheader.sequencer_loopstart then
    zsmheader.sequencer_loopend := zsmheader.sequencer_loopstart;
  if zsmheader.sequencer_loopend > zsmheader.sequencer_entries then
    zsmheader.sequencer_loopend := zsmheader.sequencer_entries;


  (* Patterns *)
  getmem(patterndata_pointer, zsmheader.patterndata_size);
  blockread(f, patterndata^, zsmheader.patterndata_size);
  inc(zsmfilesize, zsmheader.patterndata_size);

  (* Samples *)
  for i := 1 to zsmheader.sample_entries do
  begin
    getmem(sample_pointer[i], smpinfo[i].length shr 1 + smpinfo[i].length and 1);
    blockread(f, sample_pointer[i]^, smpinfo[i].length shr 1 + smpinfo[i].length and 1);
  end;

  close(f);


  skip:

  actual_spl_per_row := (2646 * qfactor) div 2;

end;


procedure adapt_zsm_pitchtables;
  var
    i, count: word;
    buf:longint;
begin

  count := 0;
  for i := 1 to zsmheader.sample_entries do
    inc(count, pitch_table_len[i]);

  for i := 0 to count -1 do
  begin
    buf := 2 * (pitch_tables^[i].vor*65536 + pitch_tables^[i].nach) div qfactor;
    pitch_tables^[i].vor := buf div 65536;
    pitch_tables^[i].nach := buf mod 65536;


  end;

end;



procedure clear_zsm;
  var i: byte;
begin

  for i := 1 to zsmheader.sample_entries do
  begin;
    freemem(sample_pointer[i], smpinfo[i].length);
    smpinfo[i].length := 0;
    smpinfo[i].loopstart := 0;
    smpinfo[i].marker := 0;
  end;

  freemem(patterndata_pointer, zsmheader.patterndata_size);
  freemem(sequencer_pointer, zsmheader.sequencer_entries * 2);
  freemem(pitch_tables_pointer, pitch_tables_bytesize);

  with zsmheader do
  begin
    fillchar(titel, 20, 0);
    sample_entries := 0;
    sequencer_entries := 0;
    sequencer_loopstart := 0;
    sequencer_loopend := 0;
    patterndata_size := 0;
  end;

end;



function readbits(length: byte): byte; assembler;
asm

  mov dl, length

  les di, patterndata_pointer

  (* 32 Bit Bitposition einlesen und auch in CX uebergeben *)
  db 66h; mov bx, word ptr bitread_pos { mov ebx, bitpos_asm }
  db 66h; mov cx, bx  { mov ecx, ebx }

  (* EBX durch 8 teilen, WORD einlesen *)
  db 66h; shr bx, 3 { shr ebx, 3 }
  mov ax, es:[di+bx]

  (* Modulo Wert erzeugen bzw. in DL steht hiernach der Bitoffset *)
  db 66h; shl bx, 3  { shl ebx, 3 }
  db 66h; sub cx, bx { sub ecx, ebx }

  (* zurechtschieben, Bitoffset elimieren *)
  shr ax, cl
  (* zu lesende Bits befinden sich nun nur noch in AL *)

  (* Maske anwenden *)
  mov cl, dl
  mov ch, 1
  shl ch, cl
  dec ch
  and al, ch

  (* bitposition weiterzaehlen *)
  db 66h; xor cx, cx
  mov cl, dl
  db 66h; add word ptr bitread_pos, cx { add bitread_pos, ecx }
end;






procedure goto_sequ(num: word);
  var
    track, i: byte;
    ii: word;
begin
  actual_sequ := num;

  bitread_pos := sequencer^[num] * 8;

  pattheaderandtables := sequencer^[num];

  offset_in_patterndata := sequencer^[num];


  (* Pattern Header, Tabellen etc. einlesen *)
  move(patterndata^[bitread_pos div 8], patternheader, 5); (* Header *)
  inc(bitread_pos, 5 * 8); (* Header eingelesen, also ueberspringen *)


  patternheader.spl_per_row := (patternheader.spl_per_row * qfactor) div 2;


  patternrowstotal := patternheader.rows;
  ii := 0;
  repeat inc(ii) until sequencer^[num] = pattern_size_table^[ii];
  act_pattern_number := ii;
  act_pattern_size := pattern_size_table^[ii+1]-pattern_size_table^[ii];

  purebits := 0;
  emptyflagbits := 0;


  (* Aktive-Tracks Flags einlesen *)
  for track := 0 to 15 do
    if patternheader.tracks_active and (1 shl track) > 0 then
      trk_active[track] := true else trk_active[track] := false;

  (* Tabellen einlesen *)
  for track := 0 to 15 do if trk_active[track] then
  begin

    if readbits(1) > 0 then
    begin

      sampnr_table_len[track] := readbits(bitsneeded[zsmheader.sample_entries+1]);
      inc(sampnr_table_len[track]);
 (*****************************************************)
      for i := 1 to sampnr_table_len[track] do
        sampnr_table[track, i-1] := readbits(bitsneeded[zsmheader.sample_entries+1]);
    end
    else
      sampnr_table_len[track] := 0;


    volume_table_len[track] := readbits(4);
    inc(volume_table_len[track]);
    for i := 1 to volume_table_len[track] do
      volume_table[track, i-1] := readbits(4);

    if readbits(1) = 0 then
      do_repeat[track] := false else do_repeat[track] := true;

    if readbits(1) = 0 then onlyvol_exist[track] := false
      else onlyvol_exist[track] := true;

  end;

  if readbits(1) = 1 then use_emptyrow_flags := true else use_emptyrow_flags := false;



  actual_row := 0;

  if patternheader.spl_per_row > 0 then
    actual_spl_per_row := patternheader.spl_per_row;
  if patternheader.spl_per_row = 0 then
    patternheader.spl_per_row := actual_spl_per_row;

  pattheaderandtables := bitread_pos div 8 - pattheaderandtables;

end;


procedure goto_next_sequ;
begin

  inc(actual_sequ);
  if actual_sequ > zsmheader.sequencer_loopend then
  begin
    actual_sequ := zsmheader.sequencer_loopstart;
    inc(loopcount);
  end;
  goto_sequ(actual_sequ);

end;



procedure read_row;
  var
    track, tblpos, i: byte;
    read_onlyvol: bytebool;
    freq: longint;
  label
    skip, skip2;
begin

  patternrow_act := actual_row;

  fillchar(sync, 256, 0);

  fillchar(smp_trig, 63, 0); { nur Player }


  if use_emptyrow_flags then inc(emptyflagbits);

  for track := 0 to 15 do
    pattern_row[track].event := false;

  {
  for track := 0 to 15 do
    npattline[track] := npattline[track] and $FFFF;
  }

  (* Leerzeilen Flag optional *)
  if use_emptyrow_flags then if readbits(1) = 0 then goto skip2;

  for track := 0 to 15 do
  if trk_active[track] then (* nur aktive Kanaele abfragen *)
  if readbits(1) = 1 then (* Wenn Event auf Kanal stattfindet *)
  begin
    pattern_row[track].event := true; (* Event markieren *)

    inc(emptyflagbits);

    (* Bei nur Lautstaerke nicht triggern *)
    read_onlyvol := false;
    if onlyvol_exist[track] then
    begin
    inc(purebits);
    if readbits(1) = 0 then
    begin
      read_onlyvol := true;
      goto skip
    end;
    end;
     fillchar(smp_position[track], 4, 0); (* triggern *)
     fillchar(smp_pos_before[track], 4, 0);
     channel_phase[track] := 0;
     channel_last_ah[track] := 0;
     channel_dh_[track] := 0{255};
     channel_last_bp[track] := 0;
     channel_trig[track] := $FF;
     channel_bx_[track] := 0;
     channel_ax_[track] := 0;
     channel_ebp[track] := 0;
     chan_trig[track] := 3;

    (* Repeat abfragen, falls genutzt *)
    if do_repeat[track] then
    begin
     inc(purebits);
     if readbits(1) = 1 then
     begin
       pattern_row[track].smp := last_pattern_row[track].smp;
       pattern_row[track].fixvor := last_pattern_row[track].fixvor;
       pattern_row[track].fixnach := last_pattern_row[track].fixnach;

       smp_trig[pattern_row[track].smp] := 3; { nur fr den Player }

       goto skip;
     end
    end;

     (* Sampnr *)
     if sampnr_table_len[track] = 0 then
     begin
       inc(purebits, bitsneeded[zsmheader.sample_entries+1]);
       pattern_row[track].smp := readbits(bitsneeded[zsmheader.sample_entries+1])
     end
     else
     begin
       inc(purebits, bitsneeded[sampnr_table_len[track]]);
       pattern_row[track].smp
         := sampnr_table[track, readbits(bitsneeded[sampnr_table_len[track]])];
     end; (* SampNr 0 zusammen mit event bedeutet Sample Stop *)


     (* Pitch *)
     if pattern_row[track].smp > 0 then
     begin
       tblpos := readbits(bitsneeded[pitch_table_len[pattern_row[track].smp]]);

       inc(purebits, bitsneeded[pitch_table_len[pattern_row[track].smp]]);

       move(pitch_tables^[pitch_table_offset[pattern_row[track].smp] + tblpos],
         pattern_row[track].fixvor, 3);

     end;

     if pattern_row[track].smp > 0 then
       move(pattern_row[track], last_pattern_row[track], 4);



     smp_trig[pattern_row[track].smp] := 3; { nur fr den Player }



     skip:

     if (pattern_row[track].smp > 0) or read_onlyvol then
     begin
       pattern_row[track].vol
        := volume_table[track, readbits(bitsneeded[volume_table_len[track]])];
       inc(purebits, bitsneeded[volume_table_len[track]]);
     end;

    play_sample_looped[track] :=  smpinfo[pattern_row[track].smp].loopstart
       < smpinfo[pattern_row[track].smp].length;

  end;

  skip2:

  for track := 0 to 15 do (* Nur fr den Player als Anzeige *)
    chan_event[track] := pattern_row[track].event;

  {
  for track := 0 to 15 do
  begin
    if pattern_row[track].event then
    begin
      freq := trunc((pattern_row[track].fixvor + pattern_row[track].fixnach/65536)*22050);
      npattline[track] := freq + $100000;
      npattline_smp[track] := pattern_row[track].smp;
    end;

    {
    if trk_active[track] and (actual_row = 0) and
    (pattern_row[track].smp > 0) and pattern_row[track].event then inc(npattline[track], 8);
    end;
    }
    {if actual_row = 0 then inc(npattline[track], 8);}
 {
   end;

  }
   for track := 0 to 15 do
   if trk_active[track] then
   if pattern_row[track].event then
     {for i := 1 to zsmheader.sample_entries do}
       sync[smpinfo[pattern_row[track].smp].marker]
         := pattern_row[track].vol;
    { nochmal berprfen! }



  {new_pattline;}

  inc(actual_row);

end;



procedure mix_in_channel_sloop; assembler;
asm

  {
  mov al, 10
  sub al, actual_chan_vol
  mov cs:[offset @voldata-1], al
  }
  {
  mov al, 10
  mov ah, actual_chan_vol
  shr ah, 1
  sub al, ah
  mov cs:[offset @voldata-1], al
  }

  mov al, 10
  mov ah, actual_chan_vol
  mov cs:[offset @voldata_full-1], ah
  shr ah, 1
  sub al, ah
  mov cs:[offset @voldata-1], al


  db 66h; mov ax, word ptr s_speed
  db 66h; mov cs:[offset @spd_data-4], ax

  mov ax, actual_smplen
  dec ax
  mov cs:[offset @act_smplen-2], ax

  mov ax, actual_sample_loopstart
  mov cs:[offset @act_smp_loopstart-2], ax


  dw 0ebh { Sprung der Lnge 0, Cache lschen }

  mov cx, ds; lds si, smp_pointer; mov ax, ds; mov ds, cx
  db 8eh; db 0e0h { mov fs, ax }

  mov si_start, si

  push bp
  db 66h; mov bp, word ptr s_pos

  db 66h; or bp, bp
  jz @skip_si
  mov si, si_
  @skip_si:


  les di, soundbuffer16bit_pointer
  mov ax, mix_position
  add ax, ax { SHL AX, 1 }
  add di, ax
  (* es:[di] zeigt nun auf die richtige Mixposition *)


  db 66h; ror di, 16
  mov di, fill_length
  dec di
  db 66h; ror di, 16

  mov dl, phase
  mov ah, last_ah
  db 66h; mov cx, word ptr s_pos_before
  mov bx, bx_
  mov dh, dh_


  jmp @skip
  @reset:



  mov bp, 07777h; {????} @act_smp_loopstart: { Sample wieder auf Loopstart setzen }

  mov si, si_start
  mov dx, bp
  shr dx, 1
  add si, dx

  mov dx, bp
  shl dx, 8
  mov dl, loopstphase

    xor bx, bx
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]

   and dh, 1;
   jnz @dontr
    add dl, bh
  mov dh, 255
  jmp @deltal_end

  @dontr:

   add dl, bl; xor bx, bx
  xor dh, dh
  @deltal_end:




  mov cx, bp


  @skip:

  @loop:
  cmp bp, 07777h; {????} @act_smplen:; jae @reset { Sample-Ende erreicht? }


 { keine neuen Werte berechnen wenn keine Bewegung }
  sub cx, bp
  jnc @skip2

(*********************************)


   and dh, 1;
   jnz @dontread

   xor dh, cl
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  inc cx; jz @deltaloop_end
    add dl, bl; xor bx, bx
  inc cx; jz @deltaloop_end
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  inc cx; jz @deltaloop_end
    add dl, bl; xor bx, bx
  inc cx; jz @deltaloop_end
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  inc cx; jz @deltaloop_end
    add dl, bl; xor bx, bx
  inc cx; jz @deltaloop_end
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  inc cx; jz @deltaloop_end

  @dontread:

  xor dh, cl
    add dl, bl; xor bx, bx
  inc cx; jz @deltaloop_end
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  inc cx; jz @deltaloop_end
    add dl, bl; xor bx, bx
  inc cx; jz @deltaloop_end
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  inc cx; jz @deltaloop_end
    add dl, bl; xor bx, bx
  inc cx; jz @deltaloop_end
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  inc cx; jz @deltaloop_end
    add dl, bl; xor bx, bx
  inc cx; jz @deltaloop_end

  @deltaloop_end:


  mov ah, dl
  mov cl, 0; @voldata_full:
  test cl, 1
  jnz @vol_odd
  mov cx, ax
  sar cx, 2
  sub ax, cx { AX * 0,75 }
  @vol_odd:
  sar ax, 0; @voldata:

  @skip2:

  add es:[di], ax

  mov cx, bp { letzte Position in cx merken }

  db 66h; db 081h, 0c5h; dw 0ffffh, 0ffffh; @spd_data:
  adc bp, 0

  db 66h; db 081h, 0c7h, 002h, 000h, 0ffh, 0ffh { add edi, 0ffff0002h }
  jc @loop


  @end:

  mov bx_, bx
  mov last_ah, ah
  mov dh_, dh
  mov phase, dl
  mov si_, si
  db 66h; mov word ptr s_pos, bp
  db 66h; mov ax, bp
  db 66h; sub ax, word ptr s_speed; sbb ax, 0
  db 66h; mov word ptr s_pos_before, ax
  pop bp

end;



procedure mix_in_channel; assembler;
asm

  push bp

  {
  mov al, 10
  sub al, actual_chan_vol
  mov cs:[offset @voldata_even-1], al
  mov cs:[offset @voldata_odd-1], al
  }
  mov al, 10
  mov ah, actual_chan_vol
  mov cs:[offset @voldata_full_even-1], ah
  mov cs:[offset @voldata_full_odd-1], ah
  shr ah, 1
  sub al, ah
  mov cs:[offset @voldata_even-1], al
  mov cs:[offset @voldata_odd-1], al

  db 66h; mov ax, word ptr s_speed;
  {mov cs:[offset @hi_wordschritt-2], ax}
  {db 66h; shr ax, 16}
{  mov cs:[offset @lo_wordschritt-2], ax}
  db 66h; mov cs:[offset @dwordschritt-4], ax

  dw 0ebh { Sprung der Lnge 0, Cache lschen }

  mov cx, ds; lds si, smp_pointer; mov ax, ds; mov ds, cx
  db 8eh; db 0e0h { mov fs, ax }

  les di, soundbuffer16bit_pointer
  mov ax, mix_position
  add ax, ax { shl ax, 1 }; add di, ax
  (* es:[di] zeigt nun auf die richtige Mixposition *)
  (* fs:[si] zeigt auf Sampledaten *)

   mov di_, di

  db 66h; mov bp, word ptr s_pos
  db 66h; mov dx, word ptr s_speed

  mov ax, actual_smplen

  cmp bp, ax
  jnb @end

  db 66h; or dx, dx
  jz @end


  mov play_sample_actual, 0

  sub ax, bp (* Rest von abzuspielendem Sample berechnen *)
  dec ax
  db 66h; shl ax, 16
  db 66h; mov bx, dx
  db 66h; xor dx, dx
  db 66h; ror bx, 16
  db 66h; div bx

  (* in EAX steht nun die Laenge, *)
  (* die verarbeitet werden kann, *)
  (* ohne das Sample-Ende zu ueberschreiten *)

{ DX kann ab hier fuer anderes verwendet werden }


  cmp trig, 255
  je @trigger
  mov si, si_
  @trigger:
  mov trig, 0


  db 66h; or ax, ax (* Wenn EAX 0 dann nichts zu berechnen *)
  jz @end

  db 66h; xor bx, bx
  mov bx, fill_length

  db 66h; cmp ax, bx
  jbe @below
  mov ax, bx
  mov play_sample_actual, 0ffh
  @below:


  db 66h; ror di, 16
  mov di, ax { EDI auf fill_length - 1 setzen }
  dec di
  db 66h; ror di, 16



  db 66h; mov bp, word ptr ebp_

  mov dl, phase
  mov bx, bx_
  mov ax, ax_


  jmp @skip_table
  @jumptable:
    dw offset @skip2even
    dw offset @delta1even
    dw offset @delta2even
    dw offset @delta3even
    dw offset @delta4even
    dw offset @delta5even
    dw offset @delta6even
    dw offset @delta7even
    dw offset @skip2odd
    dw offset @delta1odd
    dw offset @delta2odd
    dw offset @delta3odd
    dw offset @delta4odd
    dw offset @delta5odd
    dw offset @delta6odd
    dw offset @delta7odd
  @skip_table:


  @loop:
  db 66h; db 081h, 0c5h; dw 07777h, 07777h; @dwordschritt: { add ebp, ???????? }
  adc bp, 0
  jz @skip2even { fngt nur even-Durchlufe ab weil bei odd BP mit 8 startet }

  add bp, bp
  mov bp, cs:[offset @jumptable+bp]
  jmp bp

  @delta7odd:
    add dl, bl; xor bx, bx
  @delta6even:
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  @delta5odd:
    add dl, bl; xor bx, bx
  @delta4even:
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  @delta3odd:
    add dl, bl; xor bx, bx
  @delta2even:
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  @delta1odd:
    add dl, bl; xor bx, bx


  mov ah, dl
  mov cl, 0; @voldata_full_even:
  test cl, 1
  jnz @vol_odd
  mov cx, ax
  sar cx, 2
  sub ax, cx { AX * 0,75 }
  { 07 07 06 06 05 05 04 04 03 03 02 02 01 01 00 00 }
  { 15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00 }
  @vol_odd: { 15, 13, 11, ... }
  sar ax, 0; @voldata_even:


  @skip2even:

  xor bp, bp

  add es:[di], ax

  db 66h; db 081h, 0c7h, 002h, 000h, 0ffh, 0ffh { add edi, 0ffff0002h }
  jc @loop
  jmp @end


  @delta7even:
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  @delta6odd:
    add dl, bl; xor bx, bx
  @delta5even:
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  @delta4odd:
    add dl, bl; xor bx, bx
  @delta3even:
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh
  @delta2odd:
    add dl, bl; xor bx, bx
  @delta1even:
    db 64h; mov bl, ds:[si]; inc si
    add bx, bx; mov bx, ds:[bx + offset deltatab]
    add dl, bh



  mov ah, dl
  mov cl, 0; @voldata_full_odd:
  test cl, 1
  jnz @vol_odd2
  mov cx, ax
  sar cx, 2
  sub ax, cx { AX * 0,75 }
  @vol_odd2: { 15, 13, 11, ... }
  sar ax, 0; @voldata_odd:

  @skip2odd:

  mov bp, 8

  add es:[di], ax

  db 66h; db 081h, 0c7h, 002h, 000h, 0ffh, 0ffh { add edi, 0ffff0002h }
  jc @loop




  @end:
  mov ax_, ax
  mov bx_, bx
  mov phase, dl



  db 66h; xor dx, dx
  mov dx, di; sub dx, di_; shr dx, 1
  db 66h; mov ax, word ptr s_speed
  db 66h; ror ax, 16; db 66h; mul dx
  db 66h; ror ax, 16
  db 66h; add word ptr s_pos, ax; adc word ptr s_pos, 0

  db 66h; mov word ptr ebp_, bp
  pop bp

  mov si_, si

end;




procedure mix_in_channels;
  var
    track, fx: byte;
begin

  fillchar(smp_playing, 63, 0); {smp_playing nur fr player-anzeige!}
  fillchar(smp_pos_show, 126, 0);
  fillchar(channel_smp, 16, 0);
  fillchar(channel_vol, 16, 0);
  fillchar(smp_instances, 63, 0);

  for track := 0 to 15 do
  begin
  if play_sample[track] and track_select[track] then
  begin
    smp_playing[pattern_row[track].smp] := pattern_row[track].vol;
    inc(smp_instances[pattern_row[track].smp]);
    smp_freq_fine[pattern_row[track].smp] := trunc(
    11025 * word(qfactor) * (pattern_row[track].fixvor
    +pattern_row[track].fixnach / 65536));

    smp_freq[pattern_row[track].smp] := round(
      smp_freq_fine[pattern_row[track].smp] / 1000);
    {
    smp_freq[pattern_row[track].smp] := round(
      11.025 * qfactor * (pattern_row[track].fixvor + pattern_row[track].fixnach / 65536));
    }
    channel_smp[track] := pattern_row[track].smp;
    channel_vol[track] := pattern_row[track].vol;

    play_sample_actual := true;

    smp_pointer := sample_pointer[pattern_row[track].smp];

    s_pos_before := smp_pos_before[track];
    s_pos := smp_position[track];
    s_speed.vor := pattern_row[track].fixvor;
    s_speed.nach := pattern_row[track].fixnach;

    actual_smplen := smpinfo[pattern_row[track].smp].length;
    actual_chan_vol := pattern_row[track].vol;

    actual_sample_loopstart := smpinfo[pattern_row[track].smp].loopstart;

    phase := channel_phase[track];

    dh_ := channel_dh_[track];

    si_ := channel_si_[track];

    last_ah := channel_last_ah[track];

    last_bp := channel_last_bp[track];

    trig := channel_trig[track];

    bx_ := channel_bx_[track];

    ax_ := channel_ax_[track];

    move(channel_ebp[track], ebp_, 4);
{    ebp_ := channel_ebp[track];}

    if play_sample_looped[track] then
    begin
      loopstphase := smpinfo[pattern_row[track].smp].loopst_phase;
      mix_in_channel_sloop;
    end
    else
      mix_in_channel;

      inc(traffic,
        trunc(fill_length * (s_speed.vor + s_speed.nach/65536)/2));

    channel_ax_[track] := ax_;

    channel_bx_[track] := bx_;

    channel_trig[track] := trig;

    channel_last_bp[track] := last_bp;

    channel_si_[track] := si_;

    channel_dh_[track] := dh_;

    channel_last_ah[track] := last_ah;

    channel_phase[track] := phase;

{    channel_ebp[track] := ebp_;}
    move(ebp_, channel_ebp[track], 4);

    smp_position[track] := s_pos;
    smp_pos_before[track] := s_pos_before;

    play_sample[track] := play_sample_actual;

    smp_pos_show[pattern_row[track].smp] := smp_position[track].vor;

  end;
    {
    if pattern_row[track].event and (pattern_row[track].smp = 0) then
    npattline[track] := 0;
    }

    {
    if not play_sample_actual then
    begin
      npattline[track] := 0;
      npattline_smp[track] := 0;
    end;
    }
  end; { for track }





  for fx := 1 to 32 do
  begin
  if play_fx[fx] then
  begin
    play_sample_actual := true;

    smp_pointer := fx_pointer[fx];

    s_pos := fx_position[fx];
    s_speed.vor :=  fx_speed[fx].vor;
    s_speed.nach := fx_speed[fx].nach;

    actual_smplen := fx_length[fx];
    actual_chan_vol := fx_vol[fx];

    actual_sample_loopstart := 0;

    mix_in_channel;

    fx_position[fx] := s_pos;

    play_fx[fx] := play_sample_actual;
  end;
  end; { for fx }




  inc(mix_position, fill_length);

end;




procedure make_8bit_buffer_vol2; assembler;
asm

  mov cx, soundbuffer_size

  les di, soundbuffer8bit_pointer

  push ds
  lds si, soundbuffer16bit_pointer


  mov dx, -256
  mov bx, 255

  @loopstart:

  mov ax, ds:[si]
  sar ax, 6



  (* positiv: wenn ah = 0 ist kein Clipping *)

  (* negativ: wenn ah + 1 = 0 ist kein Clipping *)

  or ah, ah
  jz @skip2 (* kein Clipping *)
  js @negativ (* moegl. negatives Clipping *)
  mov ax, bx (* pos. Clipping *)
  jmp @skip2
  @negativ:
  inc ah
  jz @skip
  mov ax, dx
  jmp @skip2
  @skip:
  dec ah
  @skip2:
  sar ax, 1


  add al, 128 (* zu vorzeichenlos konvertieren *)
  mov es:[di], al

  add si, 2
  inc di

  dec cx
  jnz @loopstart

  pop ds
end;


procedure make_8bit_buffer_vol1; assembler;
asm

  mov cx, soundbuffer_size

  les di, soundbuffer8bit_pointer

  push ds
  lds si, soundbuffer16bit_pointer


  mov dx, -256
  mov bx, 255

  @loopstart:

  mov ax, ds:[si]
  sar ax, 5


  or ah, ah
  jz @skip2 (* kein Clipping *)
  js @negativ (* moegl. negatives Clipping *)
  mov ax, bx (* pos. Clipping *)
  jmp @skip2
  @negativ:
  inc ah
  jz @skip
  mov ax, dx
  jmp @skip2
  @skip:
  dec ah
  @skip2:
  sar ax, 1



  add al, 128 (* zu vorzeichenlos konvertieren *)
  mov es:[di], al

  add si, 2
  inc di

  dec cx
  jnz @loopstart

  pop ds
end;

procedure make_8bit_buffer_vol0; assembler;
asm

  mov cx, soundbuffer_size

  les di, soundbuffer8bit_pointer

  push ds
  lds si, soundbuffer16bit_pointer


  mov dx, -256
  mov bx, 255

  @loopstart:

  mov ax, ds:[si]
  sar ax, 4


  or ah, ah
  jz @skip2 (* kein Clipping *)
  js @negativ (* moegl. negatives Clipping *)
  mov ax, bx (* pos. Clipping *)
  jmp @skip2
  @negativ:
  inc ah
  jz @skip
  mov ax, dx
  jmp @skip2
  @skip:
  dec ah
  @skip2:
  sar ax, 1


  add al, 128 (* zu vorzeichenlos konvertieren *)
  mov es:[di], al

  add si, 2
  inc di

  dec cx
  jnz @loopstart

  pop ds
end;


procedure get_vumeter; assembler;
asm

  les si, soundbuffer16bit_pointer

  mov cx, soundbuffer_size
  shr cx, 2
  xor bx, bx
  xor dx, dx

  @loopstart:

    mov ax, es:[si]

    cmp ax, bx
    jge @2
    neg ax
    @2:

    cmp ax, dx
    jbe @3
    mov dx, ax
    @3:

    add si, 8


  dec cx
  jnz @loopstart

  shr dx, 4

  mov vumeter, dx

end;


procedure get_scope; assembler;
asm
  mov dx, soundbuffer_size
  shr dx, 8

  mov cx, 256

  les di, scope_ptr
  push ds; lds si, soundbuffer8bit_pointer

  @loop:
  mov al, ds:[si]
  shr al, 1
  stosb
  add si, dx

  dec cx; jnz @loop

  pop ds
end;


procedure fill_soundbuffer; (* beim erstmaligen Aufruf row_countdown = 0 *)
  var
    buffercountdown: word;
    track: byte;
  label
    skip;
begin

  traffic := 0;

  fillchar(soundbuffer16bit^, soundbuffer_size shl 1, 0);

  mix_position := 0;
  buffercountdown := soundbuffer_size;

  if row_countdown > soundbuffer_size then
  begin
    fill_length := soundbuffer_size;
    mix_in_channels;
    dec(row_countdown, fill_length);
    goto skip;
  end;


  while buffercountdown >= row_countdown do
  begin
    fill_length := row_countdown;
    mix_in_channels;
    dec(buffercountdown, fill_length);
    if actual_row = patternheader.rows + 1 then goto_next_sequ;

    read_row;
    row_countdown := patternheader.spl_per_row;

    for track := 0 to 15 do (* Bestimmen ob Sample gespielt wird *)
      if pattern_row[track].event then
        if pattern_row[track].smp > 0 then
          play_sample[track] := true
        else
          play_sample[track] := false;

  end;


  if buffercountdown > 0 then
  begin
    fill_length := buffercountdown;
    mix_in_channels;
    dec(row_countdown, fill_length);
  end;


  skip:
(* Puffer gefuellt *)


  if scopemode then get_scope;
  get_vumeter;
{  if not scopemode then get_vumeter;}

  vumeter := vumeter div (1 shl mixvol);
{  if vumeter < 1 then vumeter := 1;}

  soundbuffer8bit_pointer := doublebuffer;
  if mixvol = 1 then
    make_8bit_buffer_vol1
  else
  if mixvol = 2 then
    make_8bit_buffer_vol2
  else
    make_8bit_buffer_vol0;

end;


procedure init_zsm_soundbuffer(size: word);
begin
  getmem(soundbuffer16bit_pointer, size*2); { 16 Bit Soundpuffer }
  soundbuffer_size := size;

  fillchar(channel_phase, 16, 128);

  mixvol := 1;

  getmem(scope_ptr, 256);
  fillchar(scope_ptr^, 256, 64);

  {
  0 = 0 dB
  1 = -6 dB
  2 = -12 dB
  }
end;


function patternpos: word;
begin
  patternpos
    := bitread_pos div 8 - sequencer^[actual_sequ] - pattheaderandtables;
end;


procedure calc_pattern_sizes;
  var
    i, ii: word;
    swapbuf: word;
    swapped: bytebool;
    listlen: word;
begin
  getmem(pattern_size_table_pointer, 2048);

  move(sequencer^, pattern_size_table^, zsmheader.sequencer_entries * 2);

  swapped := true;
  while swapped do

  begin
    swapped := false;
  for i := 1 to zsmheader.sequencer_entries - 1 do
    if pattern_size_table^[i] > pattern_size_table^[i+1] then
    begin
      swapped := true;
      swapbuf := pattern_size_table^[i];
      pattern_size_table^[i] := pattern_size_table^[i+1];
      pattern_size_table^[i+1] := swapbuf;
    end;

   end;

  listlen := zsmheader.sequencer_entries;


  (* gleiche aufeinanderfolgende entfernen *)

  i := 1;
  for ii := 1 to listlen do
  begin
    if pattern_size_table^[i] <> pattern_size_table^[ii] then inc(i);
    pattern_size_table^[i] := pattern_size_table^[ii];
  end;
  listlen := i;

  pattern_size_table^[listlen+1] := zsmheader.patterndata_size;

  {
  smallest_pattern: word;
  biggest_pattern: word;
  }
  number_of_patterns := listlen;

  smallest_pattern := 65535;
  for i := 1 to listlen do
    if pattern_size_table^[i+1]-pattern_size_table^[i] < smallest_pattern then
      smallest_pattern := pattern_size_table^[i+1]-pattern_size_table^[i];


  biggest_pattern := 0;
  for i := 1 to listlen do
    if pattern_size_table^[i+1]-pattern_size_table^[i] > biggest_pattern then
      biggest_pattern := pattern_size_table^[i+1]-pattern_size_table^[i];


   for i := 1 to listlen do
     writeln(pattern_size_table^[i]);

  {freemem(pattern_size_table_pointer, 2048);}
end;

procedure inc_sequ;
begin
  if actual_sequ < zsmheader.sequencer_entries then
  begin
    inc(actual_sequ);
    goto_sequ(actual_sequ);
  end;
end;


procedure dec_sequ;
begin
  if actual_sequ > 1 then
  begin
    dec(actual_sequ);
    goto_sequ(actual_sequ);
  end;
end;



procedure get_playing_time;
  var i: word;
begin
  playing_time := 0;
  for i := 1 to zsmheader.sequencer_loopend do
  begin
    sequ_time_pos^[i] := playing_time;
    goto_sequ(i);
    inc(playing_time, patternheader.spl_per_row * (longint(patternheader.rows)+1));
  end;
  actual_spl_per_row := (2646 * qfactor) div 2;
end;


procedure get_maxtracks;
  var seq_pos: word;
      track: byte;
begin

  maxtracks := 0;
  for seq_pos := 1 to zsmheader.sequencer_loopend do
  begin

    move(patterndata^[sequencer^[seq_pos]], patternheader, 5); (* Header *)
    for track := 0 to 15 do
      if patternheader.tracks_active and (1 shl track) > 0 then
        if track > maxtracks then maxtracks := track;


  end;


end;

end.