{$G+}
unit scanzsm6; { mit 4 Bit breiten Volumes und Message }

interface


procedure reset_simu;
function load_zsm(name: string): byte; (* Laedt ein ZSTM *)
function save_zsm(name: string): byte;
function save_zsm_4bitsmp(name: string): byte;
procedure goto_sequ(num: word);
procedure read_row;
procedure trunc_null_bytes_from_sample(smp: byte);
procedure goto_next_sequ;
procedure simulate_row;
procedure evaluate_scan;
procedure apply_sample_cuts;

type zsm_header = record
  kennung: longint; (* $4D54535A "ZSTM" *)
  version: byte;
  titel: array[0..29] of char;
  composer: array[0..19] of char;
  sample_entries: byte;
  sequencer_entries: word;
  sequencer_loopstart: word;
  sequencer_loopend: word;
  patterndata_size: word;
end;

type zsm_header_v3 = record
  kennung: longint; (* $4D54535A "ZSTM" *)
  version: byte;
  titel: array[0..29] of char;
  composer: array[0..19] of char;
  date_year: word;
  date_month: byte;
  date_day: byte;
  sample_entries: byte;
  sequencer_entries: word;
  sequencer_loopstart: word;
  sequencer_loopend: word;
  patterndata_size: word;
  cmsg_size: word;
end;


type pattern_header = record
  rows: byte;
  spl_per_row: word;
  tracks_active: word;
end;

type seq_struc = array[1..1024] of word;
type shortint_struc = array[0..65527] of shortint;

type smpcut_type = record
  null_trunc: word;
  play_trunc: word;
  loop_dissolved: bytebool;
end;

type smp_info_old = record
  length: word;
  loopstart: word;
  marker: byte;
end;

type smp_info = record
  length: word;
  loopstart: word;
  loopst_phase: byte;
  marker: byte;
end;


var
  cmsg_poi: pointer;
  sequencer_pointer: pointer;
  sequencer: ^seq_struc absolute sequencer_pointer;
  chan_event: array[0..15] of bytebool;
  ptablesize: word;
  smpblocksize: longint;

  sample_pointer: array[1..63] of pointer;
  smp_pointer: pointer;
  onesample: ^shortint_struc absolute smp_pointer;
  smpcut: array[1..63] of smpcut_type;
  smpcut_trunc_total: longint;
  smpcut_play_total: longint;
  smpinfo: array[0..63] of smp_info;
  smpinfo_old: array[0..63] of smp_info_old;
  old_smp_len: array[1..63] of word;

  actual_sequ: word;
  actual_row: byte;
  zsmheader: zsm_header;
  zsmheader_new: zsm_header_v3;

  offset_in_patterndata: word;

  {
  act_pattern_number: word;
  act_pattern_size: word;
  }
  patternheader: pattern_header;

  actual_spl_per_row: word;

  zsmfilesize: longint;

  play_sample: array[0..15] of bytebool;


implementation


type fixkomma = record
  vor: byte;
  nach: word;
end;

type fixkomma_32bit = record
  vor: word;
  nach: word;
end;

type fixkomma_48bit = record
  vor: longint;
  nach: word;
end;

type pitch_table = array[0..32 * 63 - 1] of fixkomma;

type byte_struc = array[0..65527] of byte;

type integer_struc = array[0..32766] of word;


type pattern_event = record
  smp: byte;
  fixvor: byte;
  fixnach: word;
  vol: byte;
  event: bytebool;
  trigger: bytebool;
end;

type repeat_data = record
  smp: byte;
  fixvor: byte;
  fixnach: word;
end;


var
  pitch_table_len: array [1..63] of byte;
  pitch_table_offset: array [1..63] of word;
  pitch_tables_pointer: pointer;
  pitch_tables: ^pitch_table absolute pitch_tables_pointer;
  pitch_tables_bytesize: word;
  patterndata_pointer: pointer;
  patterndata: ^byte_struc absolute patterndata_pointer;
  bitread_pos: longint;
  onlyvol_exist: array[0..15] of bytebool;

  trk_active: array[0..15] of bytebool;

  sampnr_table_len: array[0..15] of byte;
  sampnr_table: array[0..15, 0..15] of byte;
  volume_table_len: array[0..15] of byte;
  volume_table: array[0..15, 0..15] of byte;
  do_repeat: array[0..15] of bytebool;
  use_emptyrow_flags: bytebool;
  row: byte;
  pattern_row: array[0..15] of pattern_event;
  last_pattern_row: array[0..15] of repeat_data;
  s_pos: fixkomma_32bit;
  s_speed: fixkomma_32bit;
  smp_position: array[0..15] of fixkomma_32bit;
  measure_smp_pos: array[1..63] of fixkomma_48bit;
  biggest_smp_pos: array[1..63] of longint;
  actual_smplen: word;
  play_sample_actual: bytebool;
  actual_chan_vol: byte;
  actual_sample_loopstart: word;
  fill_length: word;

  play_sample_looped: array[0..15] of bytebool;

  spl_pos_in_row: word;

  row_countdown: word;
  mix_position: word;


  const
    bitsneeded: array[0..256] of byte =
      (0,0,1,2,2,3,3,3,3,4,4,4,4,4,4,4,4,5,5,5,5,
      5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,
      6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
      6,6,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
      7,7,7,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
      8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
      8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
      8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
      8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
      8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
      8,8,8,8,8);

procedure apply_sample_cuts;
  var
    smp: byte;
begin

  for smp := 1 to zsmheader.sample_entries do
    if biggest_smp_pos[smp] < smpinfo[smp].length then
    begin
      smpinfo[smp].length := biggest_smp_pos[smp];
      smpinfo[smp].loopstart := smpinfo[smp].length;
    end;

end;


procedure trunc_null_bytes_from_sample(smp: byte);
  var
    i: word;
begin
  if smpinfo[smp].length < 2 then exit;
  i := smpinfo[smp].length - 1;
  smp_pointer := sample_pointer[smp];
  if smpinfo[smp].loopstart = smpinfo[smp].length
    then { nur wenn keine Schleife }
      while (i > 1) and (abs(onesample^[i]) < 2{!}) do
      begin
        dec(i);
        inc(smpcut[smp].null_trunc);
        dec(smpinfo[smp].length);
      end; { tatschlich auch Werte < +/- 2 abschneiden }
  inc(smpcut_trunc_total, smpcut[smp].null_trunc);
end;


procedure reset_simu;
begin
  fillchar(biggest_smp_pos, 4*63, 0);
  fillchar(smpcut, 63*5, 0);
  smpcut_trunc_total := 0;
  smpcut_play_total := 0;
end;



function load_zsm(name: string): byte;
var
  f: file;
  i: byte;
  {s: string;}  (* nur zum Test *)
  counter: word;
label skip;
begin

  load_zsm := 0;
  assign(f, name);
  {$I-}
  reset(f, 1);
  {$I+}
  if ioresult <> 0 then
  begin
    load_zsm := 1;
    goto skip;
  end;

  zsmfilesize := 0;

  (* Header *)
  blockread(f, zsmheader, sizeof(zsmheader));
  if zsmheader.kennung <> $4D54535A then
  begin
    close(f);
    load_zsm := 2;
    goto skip;
  end;
  inc(zsmfilesize, sizeof(zsmheader));


  (* Sample Laengen, Schleifen, Marker *)
  fillchar(smpinfo[0], 5, 0);
  blockread(f, smpinfo_old[1], zsmheader.sample_entries * 5);
  inc(zsmfilesize, zsmheader.sample_entries * 5);

  for i := 1 to zsmheader.sample_entries do
  begin
     smpinfo[i].length := smpinfo_old[i].length;
     smpinfo[i].loopstart := smpinfo_old[i].loopstart;
     smpinfo[i].marker := smpinfo_old[i].marker;
  end;

  smpblocksize := 0;
  for i := 1 to zsmheader.sample_entries do
    inc(smpblocksize, smpinfo[i].length);
  inc(zsmfilesize, smpblocksize);

  (* Pitch Tabellen *)
  blockread(f, pitch_table_len, zsmheader.sample_entries);
  inc(zsmfilesize, zsmheader.sample_entries);
  counter := 0;
  for i := 1 to zsmheader.sample_entries do
  begin
    pitch_table_offset[i] := counter div 3;
    inc(counter, pitch_table_len[i] * 3);
  end;
  pitch_tables_bytesize := counter;
  ptablesize := counter;
  getmem(pitch_tables_pointer, counter);
  blockread(f, pitch_tables^, counter);
  inc(zsmfilesize, counter);

  (* Sequencer *)
  getmem(sequencer_pointer, zsmheader.sequencer_entries * 2);
  blockread(f, sequencer^, zsmheader.sequencer_entries * 2);
  inc(zsmfilesize, zsmheader.sequencer_entries * 2);

  if zsmheader.sequencer_loopstart < 1 then zsmheader.sequencer_loopstart
    := 1;
  if zsmheader.sequencer_loopstart > zsmheader.sequencer_entries then
    zsmheader.sequencer_loopstart := zsmheader.sequencer_entries;

  if zsmheader.sequencer_loopend < zsmheader.sequencer_loopstart then
    zsmheader.sequencer_loopend := zsmheader.sequencer_loopstart;
  if zsmheader.sequencer_loopend > zsmheader.sequencer_entries then
    zsmheader.sequencer_loopend := zsmheader.sequencer_entries;


  (* Patterns *)
  getmem(patterndata_pointer, zsmheader.patterndata_size);
  blockread(f, patterndata^, zsmheader.patterndata_size);
  inc(zsmfilesize, zsmheader.patterndata_size);

  (* Samples *)
  for i := 1 to zsmheader.sample_entries do
  begin
    old_smp_len[i] := smpinfo[i].length;
    getmem(sample_pointer[i], smpinfo[i].length);
    blockread(f, sample_pointer[i]^, smpinfo[i].length);
  end;

  close(f);


  skip:

  actual_spl_per_row := 2646;

end;



function save_zsm(name: string): byte;
var
  f: file;
  i: byte;
  {s: string;}  (* nur zum Test *)
  counter: word;
label skip;
begin

  save_zsm := 0;
  assign(f, name);
  {$I-}
  rewrite(f, 1);
  {$I+}
  if ioresult <> 0 then
  begin
    save_zsm := 1;
    close(f);
    goto skip;
  end;




  (* Header *)
  blockwrite(f, zsmheader, sizeof(zsmheader));

  {inc(zsmfilesize, sizeof(zsmheader));}


  (* Sample Laengen, Schleifen, Marker *)
  blockwrite(f, smpinfo[1], zsmheader.sample_entries * 5);
{  inc(zsmfilesize, zsmheader.sample_entries * 5);}

  {
  smpblocksize := 0;
  for i := 1 to zsmheader.sample_entries do
    inc(smpblocksize, smpinfo[i].length);
  inc(zsmfilesize, smpblocksize);
}

  (* Pitch Tabellen *)
  blockwrite(f, pitch_table_len, zsmheader.sample_entries);
{  inc(zsmfilesize, zsmheader.sample_entries);}
  counter := 0;
  for i := 1 to zsmheader.sample_entries do
  begin
 {   pitch_table_offset[i] := counter div 3;}
    inc(counter, pitch_table_len[i] * 3);
  end;
  pitch_tables_bytesize := counter;
{  ptablesize := counter;}
{  getmem(pitch_tables_pointer, counter);}
{  blockread(f, pitch_tables^, counter);}
  blockwrite(f, pitch_tables^, counter);
{  inc(zsmfilesize, counter);}

  (* Sequencer *)
{  getmem(sequencer_pointer, zsmheader.sequencer_entries * 2);}
  blockwrite(f, sequencer^, zsmheader.sequencer_entries * 2);
{  blockread(f, sequencer^, zsmheader.sequencer_entries * 2);}
{  inc(zsmfilesize, zsmheader.sequencer_entries * 2);}

  {
  if zsmheader.sequencer_loopstart < 1 then zsmheader.sequencer_loopstart
    := 1;
  if zsmheader.sequencer_loopstart > zsmheader.sequencer_entries then
    zsmheader.sequencer_loopstart := zsmheader.sequencer_entries;

  if zsmheader.sequencer_loopend < zsmheader.sequencer_loopstart then
    zsmheader.sequencer_loopend := zsmheader.sequencer_loopstart;
  if zsmheader.sequencer_loopend > zsmheader.sequencer_entries then
    zsmheader.sequencer_loopend := zsmheader.sequencer_entries;
  }

  (* Patterns *)
  {getmem(patterndata_pointer, zsmheader.patterndata_size);}
  blockwrite(f, patterndata^[0], zsmheader.patterndata_size);
{  inc(zsmfilesize, zsmheader.patterndata_size);}

  (* Samples *)
  for i := 1 to zsmheader.sample_entries do
  begin
{    getmem(sample_pointer[i], smpinfo[i].length);}
    blockwrite(f, sample_pointer[i]^, smpinfo[i].length);
  end;

  close(f);


  skip:

end;


function save_zsm_4bitsmp(name: string): byte;
var
  f: file;
  i: byte;
  {s: string;}  (* nur zum Test *)
  counter: word;
label skip;
begin

  save_zsm_4bitsmp := 0;
  assign(f, name);
  {$I-}
  rewrite(f, 1);
  {$I+}
  if ioresult <> 0 then
  begin
    save_zsm_4bitsmp := 1;
    close(f);
    goto skip;
  end;




  (* Header *)
  blockwrite(f, zsmheader_new, sizeof(zsmheader_new));

  {inc(zsmfilesize, sizeof(zsmheader));}

  { Message }
  blockwrite(f, cmsg_poi^, zsmheader_new.cmsg_size);


  (* Sample Laengen, Schleifen, Marker *)
  blockwrite(f, smpinfo[1], zsmheader_new.sample_entries * 6);
{  inc(zsmfilesize, zsmheader.sample_entries * 5);}

  {
  smpblocksize := 0;
  for i := 1 to zsmheader.sample_entries do
    inc(smpblocksize, smpinfo[i].length);
  inc(zsmfilesize, smpblocksize);
}

  (* Pitch Tabellen *)
  blockwrite(f, pitch_table_len, zsmheader_new.sample_entries);
{  inc(zsmfilesize, zsmheader.sample_entries);}
  counter := 0;
  for i := 1 to zsmheader_new.sample_entries do
  begin
 {   pitch_table_offset[i] := counter div 3;}
    inc(counter, pitch_table_len[i] * 3);
  end;
  pitch_tables_bytesize := counter;
{  ptablesize := counter;}
{  getmem(pitch_tables_pointer, counter);}
{  blockread(f, pitch_tables^, counter);}
  blockwrite(f, pitch_tables^, counter);
{  inc(zsmfilesize, counter);}

  (* Sequencer *)
{  getmem(sequencer_pointer, zsmheader.sequencer_entries * 2);}
  blockwrite(f, sequencer^, zsmheader_new.sequencer_entries * 2);
{  blockread(f, sequencer^, zsmheader.sequencer_entries * 2);}
{  inc(zsmfilesize, zsmheader.sequencer_entries * 2);}

  {
  if zsmheader.sequencer_loopstart < 1 then zsmheader.sequencer_loopstart
    := 1;
  if zsmheader.sequencer_loopstart > zsmheader.sequencer_entries then
    zsmheader.sequencer_loopstart := zsmheader.sequencer_entries;

  if zsmheader.sequencer_loopend < zsmheader.sequencer_loopstart then
    zsmheader.sequencer_loopend := zsmheader.sequencer_loopstart;
  if zsmheader.sequencer_loopend > zsmheader.sequencer_entries then
    zsmheader.sequencer_loopend := zsmheader.sequencer_entries;
  }

  (* Patterns *)
  {getmem(patterndata_pointer, zsmheader.patterndata_size);}
  blockwrite(f, patterndata^[0], zsmheader_new.patterndata_size);
{  inc(zsmfilesize, zsmheader.patterndata_size);}

  (* Samples *)
  for i := 1 to zsmheader_new.sample_entries do
  begin
{    getmem(sample_pointer[i], smpinfo[i].length);}
    blockwrite(f, sample_pointer[i]^, smpinfo[i].length shr 1 + smpinfo[i].length and 1);
  end;

  close(f);


  skip:

end;



function readbits(length: byte): byte; assembler;
asm

  mov dl, length

  les di, patterndata_pointer

  (* 32 Bit Bitposition einlesen und auch in CX uebergeben *)
  db 66h; mov bx, word ptr bitread_pos { mov ebx, bitpos_asm }
  db 66h; mov cx, bx  { mov ecx, ebx }

  (* EBX durch 8 teilen, WORD einlesen *)
  db 66h; shr bx, 3 { shr ebx, 3 }
  mov ax, es:[di+bx]

  (* Modulo Wert erzeugen bzw. in DL steht hiernach der Bitoffset *)
  db 66h; shl bx, 3  { shl ebx, 3 }
  db 66h; sub cx, bx { sub ecx, ebx }

  (* zurechtschieben, Bitoffset elimieren *)
  shr ax, cl
  (* zu lesende Bits befinden sich nun nur noch in AL *)

  (* Maske anwenden *)
  mov cl, dl
  mov ch, 1
  shl ch, cl
  dec ch
  and al, ch

  (* bitposition weiterzaehlen *)
  db 66h; xor cx, cx
  mov cl, dl
  db 66h; add word ptr bitread_pos, cx { add bitread_pos, ecx }
end;






procedure goto_sequ(num: word);
  var
    track, i: byte;
begin
  actual_sequ := num;

  bitread_pos := sequencer^[num] * 8;

  (* Pattern Header, Tabellen etc. einlesen *)
  move(patterndata^[bitread_pos div 8], patternheader, 5); (* Header *)
  inc(bitread_pos, 5 * 8); (* Header eingelesen, also ueberspringen *)


  (* Aktive-Tracks Flags einlesen *)
  for track := 0 to 15 do
    if patternheader.tracks_active and (1 shl track) > 0 then
      trk_active[track] := true else trk_active[track] := false;

  (* Tabellen einlesen *)
  for track := 0 to 15 do if trk_active[track] then
  begin

    if readbits(1) > 0 then
    begin

      sampnr_table_len[track] := readbits(bitsneeded[zsmheader.sample_entries+1]);
      inc(sampnr_table_len[track]);
 (*****************************************************)
      for i := 1 to sampnr_table_len[track] do
        sampnr_table[track, i-1] := readbits(bitsneeded[zsmheader.sample_entries+1]);
    end
    else
      sampnr_table_len[track] := 0;


    volume_table_len[track] := readbits(4);
    inc(volume_table_len[track]);
    for i := 1 to volume_table_len[track] do
      volume_table[track, i-1] := readbits(4);

    if readbits(1) = 0 then
      do_repeat[track] := false else do_repeat[track] := true;

    if readbits(1) = 0 then onlyvol_exist[track] := false
      else onlyvol_exist[track] := true;

  end;

  if readbits(1) = 1 then use_emptyrow_flags := true else use_emptyrow_flags := false;



  actual_row := 0;

  if patternheader.spl_per_row > 0 then
    actual_spl_per_row := patternheader.spl_per_row;
  if patternheader.spl_per_row = 0 then
    patternheader.spl_per_row := actual_spl_per_row;

end;


procedure goto_next_sequ;
begin
  inc(actual_sequ);
  if actual_sequ > zsmheader.sequencer_loopend then
    actual_sequ := zsmheader.sequencer_loopstart;
  goto_sequ(actual_sequ);
end;

procedure read_row;
  var
    track, tblpos, i: byte;
    read_onlyvol: bytebool;
  label
    skip, skip2;
begin


  for track := 0 to 15 do
  begin
    pattern_row[track].event := false;
    pattern_row[track].trigger := false;
  end;

  (* Leerzeilen Flag optional *)
  if use_emptyrow_flags then if readbits(1) = 0 then goto skip2;

  for track := 0 to 15 do
  if trk_active[track] then (* nur aktive Kanaele abfragen *)
  if readbits(1) = 1 then (* Wenn Event auf Kanal stattfindet *)
  begin
    pattern_row[track].event := true; (* Event markieren *)

    (* Bei nur Lautstaerke nicht triggern *)
    read_onlyvol := false;
    if onlyvol_exist[track] then
    if readbits(1) = 0 then
    begin
      read_onlyvol := true;
      goto skip
    end;
     fillchar(smp_position[track], 4, 0); (* triggern *)
     pattern_row[track].trigger := true;

    (* Repeat abfragen, falls genutzt *)
    if do_repeat[track] then
     if readbits(1) = 1 then
     begin
       pattern_row[track].smp := last_pattern_row[track].smp;
       pattern_row[track].fixvor := last_pattern_row[track].fixvor;
       pattern_row[track].fixnach := last_pattern_row[track].fixnach;
       goto skip;
     end;

     (* Sampnr *)
     if sampnr_table_len[track] = 0 then
       pattern_row[track].smp := readbits(bitsneeded[zsmheader.sample_entries+1])
     else
       pattern_row[track].smp
         := sampnr_table[track, readbits(bitsneeded[sampnr_table_len[track]])];

     (* Pitch *)
     if pattern_row[track].smp > 0 then
     begin
       tblpos := readbits(bitsneeded[pitch_table_len[pattern_row[track].smp]]);
       move(pitch_tables^[pitch_table_offset[pattern_row[track].smp] + tblpos],
         pattern_row[track].fixvor, 3);
     end;

     if pattern_row[track].smp > 0 then
       move(pattern_row[track], last_pattern_row[track], 4);

     skip:

     if (pattern_row[track].smp > 0) or read_onlyvol then
       pattern_row[track].vol
        := volume_table[track, readbits(bitsneeded[volume_table_len[track]])];

    play_sample_looped[track] :=  smpinfo[pattern_row[track].smp].loopstart
       < smpinfo[pattern_row[track].smp].length;

  end;

  skip2:

  inc(actual_row);

end;


procedure simulate_row;
  var
    track: byte;
    {sample: byte;}
    longbuf,longbuf2: longint;
begin { Zeilenweise die jeweiligen "Fortschritte" berechnen, kein Puffer }

  for track := 0 to 15 do
  begin

    if pattern_row[track].trigger then
    begin {triggern}
      measure_smp_pos[pattern_row[track].smp].vor := 0;
      measure_smp_pos[pattern_row[track].smp].nach := 0;
      play_sample[track] := true;
    end;

    if play_sample[track] and (biggest_smp_pos[pattern_row[track].smp] < smpinfo[pattern_row[track].smp].length) then
    begin

      longbuf := pattern_row[track].fixvor;
      longbuf2 := actual_spl_per_row;
      inc(measure_smp_pos[pattern_row[track].smp].vor, longbuf * longbuf2);

      longbuf := pattern_row[track].fixnach;
      inc(measure_smp_pos[pattern_row[track].smp].vor, (longbuf * longbuf2) shr 16);
      inc(measure_smp_pos[pattern_row[track].smp].nach,(longbuf * longbuf2) mod 65536);

      if measure_smp_pos[pattern_row[track].smp].vor > biggest_smp_pos[pattern_row[track].smp] then
        biggest_smp_pos[pattern_row[track].smp] := measure_smp_pos[pattern_row[track].smp].vor;

      if biggest_smp_pos[pattern_row[track].smp] >= smpinfo[pattern_row[track].smp].length then
      begin
        biggest_smp_pos[pattern_row[track].smp] := smpinfo[pattern_row[track].smp].length;
        play_sample[track] := false;
      end;

    end;

  end;



  {

  - actual_spl_per_row
  - pattern_row[track].smp
  - pattern_row[track].fixvor  (pitch)
  - pattern_row[track].fixnach( pitch)

  - smp_position[track]

  - play_sample[track]

  - play_sample_looped[track]

  - pattern_row[track].event
  }
end;


procedure evaluate_scan;
  var
    sample: byte;
begin

  for sample := 1 to zsmheader.sample_entries do
  begin
    smpcut[sample].play_trunc
      := smpinfo[sample].length - biggest_smp_pos[sample];
    {
    if smpcut[sample].play_trunc > smpcut[sample].null_trunc then
      smpcut[sample].null_trunc := 0;
    }
    if smpinfo[sample].loopstart < smpinfo[sample].length then
      if smpinfo[sample].length - biggest_smp_pos[sample] >= 0 then
        smpcut[sample].loop_dissolved := true;
    inc(smpcut_play_total, smpcut[sample].play_trunc);
  end;

end;



end.